/*
 * This source file is part of the bstring string library.  This code was
 * written by Paul Hsieh in 2002-2003, and is covered by the BSD open source 
 * license. Refer to the accompanying documentation for details on usage and 
 * license.
 */

/*
 * bstraux.c
 *
 * This file is not necessarily part of the core bstring library itself, but
 * is just an auxilliary module which includes miscellaneous or trivial 
 * functions.
 */

#include <stdio.h>
#include <limits.h>
#include "bstrlib.h"
#include "bstraux.h"

#define   upcase(c) ((c) + (('a' <= (c) && (c) <= 'z') ? ('A'-'a') : 0))
#define downcase(c) ((c) + (('A' <= (c) && (c) <= 'Z') ? ('a'-'A') : 0))

/*  int bUppercase (bstring b)
 *
 *  Convert contents of bstring to upper case.
 */
int bUppercase (bstring b) {
int i, len;
	if (b == NULL) return -__LINE__;
	for (i=0, len = b->slen; i < len; i++) {
		b->data[i] = (unsigned char) upcase (b->data[i]);
	}
	return 0;
}

/*  int bLowercase (bstring b)
 *
 *  Convert contents of bstring to lower case.
 */
int bLowercase (bstring b) {
int i, len;
	if (b == NULL) return -__LINE__;
	for (i=0, len = b->slen; i < len; i++) {
		b->data[i] = (unsigned char) downcase (b->data[i]);
	}
	return 0;
}

/*  int bCaselessCmp (const bstring b0, const bstring b1)
 *
 *  Compare two strings without differentiating between case.
 */
int bCaselessCmp (const bstring b0, const bstring b1) {
int i, v, n;

	if (b0 == NULL || b1 == NULL) return SHRT_MIN;
	n = b0->slen; if (n > b1->slen) n = b1->slen;
	if (b0->slen == b1->slen && b0->data == b1->data) return 0;

	for (i = 0; i < n; i ++) {
		v = ((char) upcase (b0->data[i])) - ((char) upcase (b1->data[i]));
		if (v != 0) return v;
		if (b0->data[i] == '\0') return 0;
	}

	if (b0->slen > n) return 1;
	if (b1->slen > n) return -1;
	return 0;
}

/*  int bTrunc (bstring b, int n)
 *
 *  Truncate the bstring to at most n characters.
 */
int bTrunc (bstring b, int n) {
	if (b == NULL || n < 0) return -__LINE__;
	if (b->slen > n) {
		b->slen = n;
		b->data[n] = '\0';	/* Required for Clib interoperability */
	}
	return 0;
}

/*  bstring bTail (bstring b, int n)
 *
 *  Return with a string of the last n characters of b.
 */
bstring bTail (bstring b, int n) {
	if (b == NULL || n < 0) return NULL;
	if (n >= b->slen) return bstrcpy (b);
	return bmidstr (b, b->slen - n, n);
}

/*  bstring bHead (bstring b, int n)
 *
 *  Return with a string of the first n characters of b.
 */
bstring bHead (bstring b, int n) {
	if (b == NULL || n < 0) return NULL;
	if (n >= b->slen) return bstrcpy (b);
	return bmidstr (b, 0, n);
}

/*  bstring bTail (bstring b, int n)
 *
 *  Return with a string of the last n characters of b.
 */
int bAssign (bstring a, const bstring b) {
	if (a == NULL || b == NULL) return -__LINE__;
	a->slen = 0;
	return bconcat (a, b);
}

/*  int bFill (bstring a, char c, int len)
 *
 *  Fill a given bstring with the character in parameter c, for a length n.
 */
int bFill (bstring a, char c, int len) {
	if (a == NULL || len < 0) return -__LINE__;
	a->slen = 0;
	return bsetstr (a, len, NULL, c);
}

/*  int bSubs (bstring b, int pos, int len, const bstring a, char c)
 *
 *  Replace the substring starting at position pos, and running for a length
 *  of len with the bstring a.  If pos exceeds the length of b, then it is
 *  filled with the character c.
 */
int bSubs (bstring b, int pos, int len, const bstring a, char c) {
struct tagbstring t;

	if (a == NULL || b == NULL || (pos | len) < 0) return -__LINE__;
	if (pos >= b->slen) return bsetstr (b, pos, a, c);
	if (pos + len > b->slen) len = b->slen - pos;

	if (len >= a->slen) {
		bsetstr (b, pos, a, c);
		if (len > a->slen) bdelete (b, pos + a->slen, len - a->slen);
		return 0;
	}

	blk2tbstr (t, a->data, len);
	bsetstr (b, pos, &t, c);
	blk2tbstr (t, a->data + len, a->slen - len);
	return binsert (b, pos + len, &t, c);
}

/*  int bReplicate (bstring b, int n)
 *
 *  Replicate the contents of b end to end n times and replace it in b.
 */
int bReplicate (bstring b, int n) {
	return bpattern (b, n * b->slen);
}

/*  int bReverse (bstring b)
 *
 *  Reverse the contents of b in place.
 */
int bReverse (bstring b) {
int i, n, m;
unsigned char t;

	if (b == NULL || b->slen < 2) return -__LINE__;
	n = b->slen;
	m = ((unsigned)n) >> 1;
	n--;
	for (i=0; i < m; i++) {
		t = b->data[n - i];
		b->data[n - i] = b->data[i];
		b->data[i] = t;
	}
	return 0;
}

/*  int bStrchr (bstring b, int c)
 *
 *  Search for the character c in b.
 */
int bStrchr (bstring b, int c) {
int i, n;
unsigned char t = c;

	if (b == NULL) return -__LINE__;
	n = b->slen;
	for (i=0; i < n; i++) if (b->data[i] == t) return i;
	return -1;
}
