/*
 * This source file is part of the bstring string library.  This code was
 * written by Paul Hsieh in 2002-2003, and is covered by the BSD open source 
 * license. Refer to the accompanying documentation for details on usage and 
 * license.
 */

/*
 * bstrlib.c
 *
 * This file is the core module for implementing the bstring functions.
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "bstrlib.h"

/* Optionally include a mechanism for debugging memory */

#ifdef MEMORY_DEBUG
#include "memdbg.h"
#endif

/* Just a wrapper for memcpy.  Use a faster assembly alternative if the 
   compiler is WATCOM C/C++ */

#define bBlockCopy(D,S,L) { if ((L) > 0) memmove ((D),(S),(L)); }

/* Compute the snapped size for a given requested size.  By snapping to powers
   of 2 like this, repeated reallocations are avoided. */
static int snapUpSize (int i) {
	if (i < 8) {
		i = 8;
	} else {
		unsigned int j;
		j = i;

		/* Assumes your system is at least 32 bits, and your string is
		   at most 4GB is size. */
		j |= (j >>  1);
		j |= (j >>  2);
		j |= (j >>  4);
		j |= (j >>  8);
		j |= (j >> 16);
		i = j + 1;		/* Least power of two greater than i */
	}
	return i;
}

/*  int balloc (bstring s, int len)
 *
 *  Increase the size of the memory backing the bstring s to at least len.
 */
int balloc (bstring s, int len) {
	if (s == NULL || s->slen < 0 || s->mlen < 0 || s->mlen < s->slen || 
	    len <= 0) {
		return BSTR_ERR;
	}

	if (len > s->mlen) {
		unsigned char * x;

		if (s->mlen < s->slen + 8) {

			/* If slen is close to mlen in size then use realloc to reduce
			   the memory defragmentation */

			len = snapUpSize (len);
			x = (unsigned char *)realloc (s->data, len);
			if (x == NULL) return BSTR_ERR;
		} else {

			/* If slen is not close to mlen then avoid the penalty of copying
			   the extra bytes that are allocated, but not considered part of
			   the string */

			len = snapUpSize (len);
			x = (unsigned char *)malloc (len);
			if (x == NULL) return BSTR_ERR;
			bBlockCopy ((char *) x, (char *) s->data, s->slen);
			free (s->data);
		}
		s->data = x;
		s->mlen = len;
		s->data[s->slen] = '\0';
	}

	return BSTR_OK;
}

/*  bstring cstr2bstr (const char * str)
 *
 *  Create a bstring which contains the content of the Clib-string str.
 */
bstring cstr2bstr (const char * str) {
bstring b;
int i,j;

	if (str == NULL) return NULL;
	b = (bstring) malloc (sizeof (struct tagbstring));
	if (b == NULL) return NULL;
	j = (strlen) (str); 
	b->slen = j;

	i = j + (2 - (j != 0));
	i = snapUpSize (i);

	b->mlen = i;

	b->data = (unsigned char *) malloc (b->mlen) ;
	if (b->data == NULL) {
		free (b);
		return NULL;
	}

	for (i=0; i <= j; i ++) {
		b->data[i] = str[i];
	}

	return b;
}

/*  bstring blk2bstr (const void * blk, int len)
 *
 *  Create a bstring which contains the content of the block blk of length 
 *  len.
 */
bstring blk2bstr (const void * blk, int len) {
bstring b;
int i;

	if (blk == NULL || len < 0) return NULL;
	b = (bstring) malloc (sizeof (struct tagbstring));
	if (b == NULL) return NULL;
	b->slen = len;

	i = len + (2 - (len != 0));
	i = snapUpSize (i);

	b->mlen = i;

	b->data = (unsigned char *) malloc (b->mlen) ;
	if (b->data == NULL) {
		free (b);
		return NULL;
	}

	for (i=0; i < len; i ++) {
		b->data[i] = ((char *) blk)[i];
	}
	b->data[len] = '\0';

	return b;
}

/*  char * bstr2cstr (const bstring s, char z)
 *
 *  Create a Clib-string which is equal to the contents of the bstring s, 
 *  except that any contained '\0' characters are converted to the character 
 *  in z. This returned value should be freed with free() call, by the calling 
 *  application.
 */
char * bstr2cstr (const bstring s, char z) {
int i,l;
char * r;

	l = s->slen;
	r = (char *)malloc (l + 1);
	if (r == NULL) return r;

	for (i=0; i < l; i ++) {
		r[i] = (s->data[i] == 0) ? z : (char) (s->data[i]);
	}

	r[l] = '\0';

	return r;
}

/*  int bconcat (bstring b0, const bstring b1)
 *
 *  Concatenate the bstring b1 to the bstring b0.
 */
int bconcat (bstring b0, const bstring b1) {
int len, d;

	if (b0 == NULL || b1 == NULL) return BSTR_ERR;
	d = b0->slen;
	len = b1->slen;
	if ((d | (b0->mlen - d) | len) < 0) return BSTR_ERR;
	if (balloc (b0, d + len + 1) != 0) return BSTR_ERR;
	bBlockCopy (&b0->data[d], &b1->data[0], len);
	b0->data[d + len] = '\0';
	b0->slen += len;
	return 0;
}

/*  int bconchar (bstring b0, char c)
 *
 *  Concatenate the single character c to the bstring b0.
 */
int bconchar (bstring b0, char c) {
int d;

	if (b0 == NULL) return BSTR_ERR;
	d = b0->slen;
	if ((d | (b0->mlen - d)) < 0 || balloc (b0, d + 2) != 0) return BSTR_ERR;
	b0->data[d] = (unsigned char) c;
	b0->data[d + 1] = '\0';
	b0->slen++;
	return 0;
}

/*  bstring bstrcpy (const bstring b)
 *
 *  Create a copy of the bstring b.
 */
bstring bstrcpy (const bstring b) {
bstring b0;
int i,j;

	if (b == NULL) {
		/* Attempted to copy a NULL string */
		return b;
	}

	b0 = (bstring) malloc (sizeof (struct tagbstring));
	if (b0 == NULL) {
		/* Unable to allocate memory for string header */
		return NULL;
	}

	i = b->slen;
	j = snapUpSize (i + 1);

	b0->data = (unsigned char *) malloc (j);
	if (b0->data == NULL) {
		j = i + 1;
		b0->data = (unsigned char *) malloc (j);
		if (b0->data == NULL) {
			/* Unable to allocate memory for string data */
			free (b0);
			return NULL;
		}
	}

	b0->mlen = j;
	b0->slen = i;

	bBlockCopy ((char *) b0->data, (char *) b->data, i);
	b0->data[b0->slen] = '\0';

	return b0;
}

/*  int biseq (const bstring b0, const bstring b1)
 *
 *  Compare the string b0 and b1.  If the strings differ, 0 is returned, if 
 *  the strings are the same, 1 is returned, if there is an error, -1 is 
 *  returned.  If the length of the strings are different, this function is
 *  O(1).
 */
int biseq (const bstring b0, const bstring b1) {
int i, j;

	if (b0 == NULL || b1 == NULL) return BSTR_ERR;
	if (b0->slen != b1->slen) return 0;
	if (b0->data == b1->data) return 1;
	j = b0->slen;
	for (i = 0; i < j; i ++) {
		if (b0->data[i] != b1->data[i]) return 0;
	}
	return 1;
}

/*  int bstrcmp (const bstring b0, const bstring b1)
 *
 *  Compare the string b0 and b1.  If there is an error, SHRT_MIN is returned, 
 *  otherwise a value less than or greater than zero, indicating that the 
 *  string pointed to by b0 is lexicographically less than or greater than 
 *  the string pointed to by b1 is returned.  If the the string lengths are 
 *  unequal but the characters up until the length of the shorter are equale 
 *  then a value less than, or greater than zero, indicating that the string 
 *  pointed to by b0 is shorter or longer than the string pointed to by b1 is 
 *  returned.  0 is returned if and only if the two strings are the same.  If 
 *  the length of the strings are different, this function is O(n).
 */
int bstrcmp (const bstring b0, const bstring b1) {
int i, v, n;

	if (b0 == NULL || b1 == NULL) return SHRT_MIN;
	n = b0->slen; if (n > b1->slen) n = b1->slen;
	if (b0->slen == b1->slen && b0->data == b1->data) return 0;

	for (i = 0; i < n; i ++) {
		v = ((char) b0->data[i]) - ((char) b1->data[i]);
		if (v != 0) return v;
		if (b0->data[i] == '\0') return 0;
	}

	if (b0->slen > n) return 1;
	if (b1->slen > n) return -1;
	return 0;
}

/*  int bstrncmp (const bstring b0, const bstring b1, int n)
 *
 *  Compare the string b0 and b1 for at most n characters.  If there is an 
 *  error, SHRT_MIN is returned, otherwise a value is returned as if b0 and 
 *  b1 were first truncated to at most n characters then bstrcmp was called
 *  with these new strings are paremeters.  If the length of the strings are 
 *  different, this function is O(n).
 */
int bstrncmp (const bstring b0, const bstring b1, int n) {
int i, v, m;

	if (b0 == NULL || b1 == NULL) return SHRT_MIN;
        m = n;
	if (m > b0->slen) m = b0->slen;
	if (m > b1->slen) m = b1->slen;

	if (b0->data != b1->data) {
		for (i = 0; i < m; i ++) {
			v = ((char) b0->data[i]) - ((char) b1->data[i]);
			if (v != 0) return v;
			if (b0->data[i] == '\0') return 0;
		}
	}

	if (n == m) return 0;

	if (b0->slen > m) return 1;
	return -1;
}

/*  bstring bmidstr (const bstring b, int left, int len)
 *
 *  Create a bstring which is the substring of b starting from position left
 *  and running for a length len (clamped by the end of the bstring b.)
 */
bstring bmidstr (const bstring b, int left, int len) {
struct tagbstring b1;

	if (len > b->slen - left) len = b->slen - left;

	if (len <= 0) {
		return cstr2bstr ("");
	}

	b1 = *b;

	b1.data += left;
	b1.slen = len;

	return bstrcpy (&b1);
}

/*  void bdestroy (bstring b)
 *
 *  free up the bstring.
 */
void bdestroy (bstring b) {
	if (b == NULL || b->slen < 0 || b->mlen <= 0 || b->data == NULL)
		return;

	free (b->data);

	/* In case there is any stale usage, there is one more chance to 
	   notice this error. */

	b->slen = -1;
	b->mlen = -1;
	b->data = NULL;

	free (b);
}

/*  int binstr (const bstring s1, int pos, const bstring s2)
 *
 *  Search for the bstring s2 in s1 starting from position pos, and searching 
 *  forward.  If it is found then return with the first position where it is 
 *  found, otherwise return BSTR_ERR.  Note that this is just a brute force 
 *  string searcher that does not attempt clever things like the Boyer-Moore 
 *  search algorithm.  Because of this there are many degenerate cases where 
 *  this can take much longer than it needs to.
 */
int binstr (const bstring s1, int pos, const bstring s2) {
int j, i, l;

	if (s1 == NULL || s2 == NULL) return BSTR_ERR;
	if (s1->slen <= pos || pos < 0) return BSTR_ERR;
	if (s2->slen == 0) return pos;

	/* Obvious alias case */
	if (s1->data == s2->data) return (pos == 0)?0:BSTR_ERR;

	l = s1->slen - s2->slen + 1;

	/* No space to find such a string? */
	if (l <= pos) return BSTR_ERR;

	i = pos;
	j = 0;

	for (;;) {
		if (s2->data[j] == s1->data[i + j]) {
			j ++;
			if (j >= s2->slen) return i;
		} else {
			i ++;
			if (i >= l) break;
			j=0;
		}
	}

	return BSTR_ERR;
}

/*  int binstrr (const bstring s1, int pos, const bstring s2)
 *
 *  Search for the bstring s2 in s1 starting from position pos, and searching 
 *  backward.  If it is found then return with the first position where it is 
 *  found, otherwise return BSTR_ERR.  Note that this is just a brute force 
 *  string searcher that does not attempt clever things like the Boyer-Moore 
 *  search algorithm.  Because of this there are many degenerate cases where 
 *  this can take much longer than it needs to.
 */
int binstrr (const bstring s1, int pos, const bstring s2) {
int j, i, l;

	if (s1 == NULL || s2 == NULL) return BSTR_ERR;
	if (s1->slen <= pos || pos < 0) return BSTR_ERR;
	if (s2->slen == 0) return pos;

	/* Obvious alias case */
	if (s1->data == s2->data) return (pos == 0)?0:BSTR_ERR;

	i = pos;
	l = s1->slen - s2->slen;
	if (l < 0) return BSTR_ERR;

	/* If no space to find such a string then snap back */
	if (l + 1 <= i) i = l;
	j = 0;

	for (;;) {
		if (s2->data[j] == s1->data[i + j]) {
			j ++;
			if (j >= s2->slen) return i;
		} else {
			i --;
			if (i < 0) break;
			j=0;
		}
	}

	return BSTR_ERR;
}

/*  int binchr (const bstring b0, int pos, const bstring b1);
 *
 *  Search for the first position in b0 starting from pos or after, in which 
 *  one of the characters in b1 is found.  If such a position does not exist
 *  in b0, then BSTR_ERR is returned.
 */
int binchr (const bstring b0, int pos, const bstring b1) {
struct { unsigned long content[256 / 32]; } chrs;
unsigned char c;
int i;

	if (pos < 0 || b0 == NULL || b1 == NULL || b1->slen <= 0 
	 || b0->slen <= pos) return BSTR_ERR;
	for (i=0; i < (256 / 32); i++) chrs.content[i] = 0x0;
	for (i=0; i < b1->slen; i++) {
		c = b1->data[i];
		chrs.content[c >> 5] |= ((long)1) << (c & 31);
	}

	for (i=pos; i < b0->slen; i++) {
		c = b0->data[i];
		if (chrs.content[c >> 5] & (((long)1) << (c & 31))) return i;
	}
	return BSTR_ERR;
}

/*  int binchrr (const bstring b0, int pos, const bstring b1);
 *
 *  Search for the last position in b0 no greater than pos, in which one of 
 *  the characters in b1 is found.  If such a position does not exist in b0, 
 *  then BSTR_ERR is returned.
 */
int binchrr (const bstring b0, int pos, const bstring b1) {
struct { unsigned long content[256 / 32]; } chrs;
unsigned char c;
int i;

	if (pos < 0 || b0 == NULL || b1 == NULL || b1->slen <= 0 
	 || b0->slen <= pos) return BSTR_ERR;
	for (i=0; i < (256 / 32); i++) chrs.content[i] = 0x0;
	for (i=0; i < b1->slen; i++) {
		c = b1->data[i];
		chrs.content[c >> 5] |= ((long)1) << (c & 31);
	}

	for (i=pos; i >= 0; i--) {
		c = b0->data[i];
		if (chrs.content[c >> 5] & (((long)1) << (c & 31))) return i;
	}
	return BSTR_ERR;
}

/*  int bsetstr (bstring b0, int pos, bstring b1, unsigned char fill)
 *
 *  Overwrite the string b0 starting at position pos with the string b1. If 
 *  the position pos is past the end of b0, then the character "fill" is 
 *  appended as necessary to make up the gap between the end of b0 and pos.
 *  If b1 is NULL, it behaves as if it were a 0-length string.
 */
int bsetstr (bstring b0, int pos, const bstring b1, unsigned char fill) {
int d;
int newlen;
	if (pos < 0 || b0 == NULL || b0->slen < 0 || b0->mlen < b0->slen) return BSTR_ERR;
	if (b1 != NULL && b1->slen < 0) return BSTR_ERR;

	/* Aliasing case */
	if ((b1 != NULL) && (b0->data == b1->data) && (pos < b0->slen)) {
		bstring tmp = bstrcpy (b1);
		if (tmp == NULL) return BSTR_ERR;
		d = bsetstr (b0, pos, tmp, fill);
		bdestroy (tmp);
		return d;
	}

	/* Increase memory size if necessary */
	d = pos + blen (b1);
	if (balloc (b0, d + 1) != 0) return BSTR_ERR;

	newlen = b0->slen;

	/* Fill in "fill" character as necessary */
	if (pos > newlen) {
		int i;
		for (i = b0->slen; i < pos; i ++) b0->data[i] = fill;
		newlen = pos;
	}

	/* Copy b1 to position pos in b0. */
	if (b1 != NULL) {
		bBlockCopy ((char *) (b0->data + pos), (char *) b1->data, b1->slen);
	}

	/* Indicate the potentially increased size of b0 */
	if (d > newlen) newlen = d;

	b0->slen = newlen;
	b0->data[newlen] = '\0';

	return BSTR_OK;
}

/*  int binsert (bstring s1, int pos, bstring s2, unsigned char fill)
 *
 *  Inserts the string s2 into s1 at position pos.  If the position pos is 
 *  past the end of s1, then the character "fill" is appended as necessary to 
 *  make up the gap between the end of s1 and pos.
 */
int binsert (bstring s1, int pos, const bstring s2, unsigned char fill) {
int d, l;

	if (pos < 0 || s1 == NULL || s2 == NULL || s1->slen < 0 || 
	    s2->slen < 0 || s1->mlen < s1->slen) return BSTR_ERR;

	/* Aliasing case */
	if ((unsigned int)(s2->data - s1->data) < (unsigned int)s1->slen) {
		bstring tmp;
		d = binsert (s1, pos, tmp = bstrcpy (s2), fill);
		bdestroy (tmp);
		return d;
	}

	/* Compute the two possible end pointers */
	d = s1->slen + s2->slen;
	l = pos + s2->slen;

	if (l > d) {
		/* Inserting past the end of the string */
		if (balloc (s1, l + 1) != 0) return BSTR_ERR;
		for (d = s1->slen; d < pos; d++) s1->data[d] = fill;
		s1->slen = l;
	} else {
		/* Inserting in the middle of the string */
		if (balloc (s1, d + 1) != 0) return BSTR_ERR;
		for (l = d - 1; l >= pos + s2->slen; l--) {
			s1->data[l] = s1->data[l - s2->slen];
		}
		s1->slen = d;
	}
	bBlockCopy (s1->data + pos, s2->data, s2->slen);
	s1->data[s1->slen] = '\0';
	return 0;
}

/*  int binsertch (bstring s1, int pos, int len, unsigned char fill)
 *
 *  Inserts the character fill repeatedly into s1 at position pos for a 
 *  length len.  If the position pos is past the end of s1, then the 
 *  character "fill" is appended as necessary to make up the gap between the 
 *  end of s1 and the position pos + len.
 */
int binsertch (bstring s1, int pos, int len, unsigned char fill) {
int d, l;

	if (pos < 0 || s1 == NULL || s1->slen < 0 || s1->mlen < s1->slen ||
	    len < 0) return BSTR_ERR;

	/* Compute the two possible end pointers */
	d = s1->slen + len;
	l = pos + len;

	if (l > d) {
		/* Inserting past the end of the string */
		if (balloc (s1, l + 1) != 0) return BSTR_ERR;
		len += pos - s1->slen;
		pos = s1->slen;
		s1->slen = l;
	} else {
		/* Inserting in the middle of the string */
		if (balloc (s1, d + 1) != 0) return BSTR_ERR;
		for (l = d - 1; l >= pos + len; l--) {
			s1->data[l] = s1->data[l - len];
		}
		s1->slen = d;
	}

	d = pos + len;
	for (l=pos; l < d; l++) s1->data[l] = fill;
	s1->data[s1->slen] = '\0';
	return 0;
}

/*  int bdelete (bstring s1, int pos, int len)
 *
 *  Removes characters from pos to pos+len-1 and shifts the tail of the 
 *  bstring starting from pos+len to pos.  len must be positive for this call
 *  to have any effect.
 */
int bdelete (bstring s1, int pos, int len) {
	if (pos < 0 || len < 0 || s1 == NULL || s1->slen < 0 || s1->mlen < s1->slen) 
		return BSTR_ERR;
	if (len > 0 && pos < s1->slen) {
		if (pos + len >= s1->slen) {
			s1->slen = pos;
		} else {
			bBlockCopy ((char *) (s1->data + pos),
			            (char *) (s1->data + pos + len), 
			            s1->slen - (pos+len));
			s1->slen -= len;
		}
		s1->data[s1->slen] = '\0';
	}
	return BSTR_OK;
}

/*  int bpattern (bstring b, int len)
 *
 *  Replicate the starting bstring, b, end to end repeatedly until it surpasses
 *  len characters, then chop the result to exactly len characters.  This 
 *  function operates in-place.  The function will return with BSTR_ERR if b 
 *  is NULL or of length 0, otherwise BSTR_OK is returned.
 */
int bpattern (bstring b, int len) {
int i, d;

	d = blen (b);
	if (d <= 0 || len < 0 || balloc (b, len + 1)) return BSTR_ERR;
	if (d == 1) return bsetstr (b, len, NULL, b->data[0]);
	for (i = d; i < len; i++) b->data[i] = b->data[i - d];
	b->data[len] = '\0';
	b->slen = len;
	return 0;
}

/*  bstring bgets (bNgetc getcPtr, void * parm, char terminator)
 *
 *  Use an fgetc-like single character stream reading function (getcPtr) to 
 *  obtain a sequence of characters which are concatenated into a bstring.  
 *  The stream read is terminated by the passed in terminator function.
 *
 *  If getcPtr returns with a negative number, or the terminator character 
 *  (which is appended) is read, then the stream reading is halted as the 
 *  result obtained thus far is returned.  If no characters are read, or 
 *  there is some other detectable error, NULL is returned.
 */
bstring bgets (bNgetc getcPtr, void * parm, char terminator) {
int c, d, e;
bstring buff;

	if (getcPtr == NULL) return NULL;
	buff = cstr2bstr ("");
	if (buff == NULL) return NULL;
	d = 0;
	e = buff->mlen - 2;

	while ((c = getcPtr (parm)) >= 0) {
		if (d > e) {
			buff->slen = d;
			if (balloc (buff, d + 2) != 0) {
				bdestroy (buff);
				return NULL;
			}
			e = buff->mlen - 2;
		}
		buff->data[d] = (unsigned char) c;
		d++;
		if (c == terminator) break;
	}

	buff->data[d] = '\0';
	buff->slen = d;

	if (d == 0) {
		bdestroy (buff);
		buff = NULL;
	}

	return buff;
}

#define BS_BUFF_SZ (1024)

/*  bstring bread (bNread readPtr, void * parm)
 *
 *  Translate a finite buffer fread-like function into something which can
 *  fill in a bstring in a roughly efficient way.
 */
bstring bread (bNread readPtr, void * parm) {
int i, l, n;
bstring buff;

	if (readPtr == NULL) return NULL;
	buff = cstr2bstr ("");
	if (buff == NULL) return NULL;

	for (i=0, n=16; ; n += ((n < BS_BUFF_SZ) ? n : BS_BUFF_SZ)) {
		balloc (buff, n + 1);
		l = readPtr ((void *) (buff->data + i), 1, n - i, parm);
		i += l;
		buff->slen = i;
		if (i < n) break;
	}

	buff->data[i] = '\0';
	return buff;
}

struct bStream {
	bstring buff;		/* Buffer for over-reads */
	void * parm;		/* The stream handle for core stream */
	bNread readFnPtr;	/* fread compatible fnptr for core stream */
	int maxBuffSz;
};

/*  struct bStream * bsopen (bNread readPtr, void * parm)
 *
 *  Wrap a given open stream (described by a fread compatible function 
 *  pointer and stream handle) into an open bStream suitable for the bstring 
 *  library streaming functions.
 */
struct bStream * bsopen (bNread readPtr, void * parm) {
struct bStream * s;

	s = (struct bStream *) malloc (sizeof (struct bStream));
	if (s == NULL) return NULL;
	s->parm = parm;
	s->buff = cstr2bstr ("");
	s->maxBuffSz = BS_BUFF_SZ;
	bsetstr (s->buff, s->maxBuffSz, NULL, '?');
	s->buff->slen = 0;
	s->readFnPtr = readPtr;
	return s;
}

/*  int bsbufflength (struct bStream * s, int sz)
 *
 *  Set the length of the buffer used by the bsStream.  If sz is zero, the 
 *  length is not set.  This function returns with the previous length.
 */
int bsbufflength (struct bStream * s, int sz) {
int oldSz;
	if (s == NULL || sz < 0) return BSTR_ERR;
	oldSz = s->maxBuffSz;
	if (sz > 0) s->maxBuffSz = sz;
	return oldSz;
}

/*  void * bsclose (struct bStream * s)
 *
 *  Close the bStream, and return the handle to the stream that was originally
 *  used to open the given stream.
 */
void * bsclose (struct bStream * s) {
void * parm;
	if (s == NULL) return NULL;
	s->readFnPtr = NULL;
	if (s->buff) bdestroy (s->buff);
	s->buff = NULL;
	parm = s->parm;
	free (s);
	return parm;
}

/*  int bsreadln (bstring r, struct bStream * s, char terminator)
 *
 *  Read a bstring terminated by the terminator character or the end of the
 *  stream from the bStream (s) and return it into the parameter r.  This 
 *  function may read additional characters from the core stream that are not 
 *  returned, but will be retained for subsequent read operations.
 */
int bsreadln (bstring r, struct bStream * s, char terminator) {
int i, l;
char * b;
struct tagbstring x;

	if (s == NULL || s->buff == NULL || r == NULL) return BSTR_ERR;
	l = s->buff->slen;
	b = (char *) s->buff->data;
	x.data = (unsigned char *) b;
	r->slen = 0;

	do {
		for (i=0; i < l; i++) {
			if (b[i] == terminator) {
				int ret;
				x.slen = i + 1;
				ret = bconcat (r, &x);
				s->buff->slen = l;
				if (BSTR_OK == ret) bdelete (s->buff, 0, i + 1);
				return BSTR_ERR & -(r->slen == 0);
			}
		}

		x.slen = l;
		if (BSTR_OK != bconcat (r, &x)) break;

		l = s->readFnPtr (b, 1, s->maxBuffSz, s->parm);

	} while (l > 0);
	if (l < 0) l = 0;
	s->buff->slen = l;
	return BSTR_ERR & -(r->slen == 0);
}

/*  int bsread (bstring r, struct bStream * s, int n)
 *
 *  Read a bstring of length n (or, if it is fewer, as many bytes as is 
 *  remaining) from the bStream.  This function may read additional 
 *  characters from the core stream that are not returned, but will be 
 *  retained for subsequent read operations.  This function will not read
 *  additional characters from the core stream.
 */
int bsread (bstring r, struct bStream * s, int n) {
int l;
char * b;
struct tagbstring x;

	if (s == NULL || s->buff == NULL || r == NULL || n <= 0) return BSTR_ERR;
	l = s->buff->slen;
	b = (char *) s->buff->data;
	x.data = (unsigned char *) b;
	r->slen = 0;

	do {
		if (l + r->slen >= n) {
			int ret;
			x.slen = n - r->slen;
			ret = bconcat (r, &x);
			s->buff->slen = l;
			if (BSTR_OK == ret) bdelete (s->buff, 0, x.slen);
			return BSTR_ERR & -(r->slen == 0);
		}

		x.slen = l;
		if (BSTR_OK != bconcat (r, &x)) break;

		l = n - r->slen;
		if (l > s->maxBuffSz) l = s->maxBuffSz;

		l = s->readFnPtr (b, 1, l, s->parm);

	} while (l > 0);
	if (l < 0) l = 0;
	s->buff->slen = l;
	return BSTR_ERR & -(r->slen == 0);
}

/*  int bsunread (struct bStream * s, const bstring b)
 *
 *  Insert a bstring into the bStream at the current position.  These 
 *  characters will be read prior to those that actually come from the core 
 *  stream.
 */
int bsunread (struct bStream * s, const bstring b) {
	if (s == NULL || s->buff == NULL) return BSTR_ERR;
	return binsert (s->buff, 0, b, '?');
}

/*  int bspeek (bstring r, struct bStream * s)
 *
 *  Return the currently buffered characters from the bStream that will be 
 *  read prior to reads from the core stream.
 */
int bspeek (bstring r, struct bStream * s) {
	if (s == NULL || s->buff == NULL || r == NULL || r->slen < r->mlen || r->slen < 0) return BSTR_ERR;
	r->slen = 0;
	return bconcat (r, s->buff);
}

#include <stddef.h>

#ifdef offsetof
#define BLE_SZ (offsetof (struct bstrList, entry))
#else
#define BLE_SZ (sizeof (struct bstrList))
#endif

/*  struct bstrList * bsplit (const bstring str, unsigned char splitChar)
 *
 *  Create an array of sequential substrings from str divided by the character
 *  splitChar.  
 */
struct bstrList * bsplit (const bstring str, unsigned char splitChar) {
int mlen;
struct bstrList * strCol;
int p, i, idx;
struct tagbstring s;

	if (str == NULL) return NULL;

	mlen   = 4;
	strCol = (struct bstrList *) malloc (BLE_SZ + sizeof (bstring) * mlen);
	p      = 0;
	idx    = 0;

	do {
		i = mlen;
		while (idx >= i) i = i + i;
		if (i != mlen) {
			struct bstrList * tmp;
			tmp = (struct bstrList *) malloc (BLE_SZ + sizeof (bstring) * i);
			memcpy (tmp, strCol, BLE_SZ + sizeof (bstring) * mlen);
			free (strCol);
			strCol = tmp;
			mlen = i;
		}

		s.data = &(str->data[p]);
		for (i=p; i < str->slen; i++) {
			if (str->data[i] == splitChar) break;
		}

		if (i > p) {
			s.mlen = 1 + (i - p);
			s.slen = (i - p);
			strCol->entry[idx] = bstrcpy (&s);
			idx ++;
		}

		p = i + 1;
	} while (p < str->slen);

	strCol->qty = idx;
	return strCol;
}

/*  struct bstrList * bsplits (const bstring str, bstring splitStr)
 *
 *  Create an array of sequential substrings from str divided by any of the 
 *  characters is splitStr.
 */
struct bstrList * bsplits (const bstring str, const bstring splitStr) {
struct { unsigned long content[256 / 32]; } chrs;
unsigned char c;
int i;
int mlen;
struct bstrList * strCol;
int p, idx;
struct tagbstring s;

	if (str == NULL || splitStr == NULL || splitStr->slen == 0) return NULL;
	if (splitStr->slen == 1) return bsplit (str, splitStr->data[0]);

	for (i=0; i < (256 / 32); i++) chrs.content[i] = 0x0;
	for (i=0; i < splitStr->slen; i++) {
		c = splitStr->data[i];
		chrs.content[c >> 5] |= ((long)1) << (c & 31);
	}

	mlen   = 4;
	strCol = (struct bstrList *) malloc (BLE_SZ + sizeof (bstring) * mlen);
	p      = 0;
	idx    = 0;

	do {
		i = mlen;
		while (idx >= i) i = i + i;
		if (i != mlen) {
			struct bstrList * tmp;
			tmp = (struct bstrList *) malloc (BLE_SZ + sizeof (bstring) * i);
			memcpy (tmp, strCol, BLE_SZ + sizeof (bstring) * mlen);
			free (strCol);
			strCol = tmp;
			mlen = i;
		}

		s.data = &(str->data[p]);
		for (i=p; i < str->slen; i++) {
			c = str->data[i];
			if (chrs.content[c >> 5] & ((long)1) << (c & 31)) break;
		}

		if (i > p) {
			s.mlen = 1 + (i - p);
			s.slen = (i - p);
			strCol->entry[idx] = bstrcpy (&s);
			idx ++;
		}

		p = i + 1;
	} while (p < str->slen);

	strCol->qty = idx;
	return strCol;
}

/*  int bstrListDestroy (struct bstrList * sl)
 *
 *  Destroy a bstrList that has been created by bsplit or bsplits.
 */
int bstrListDestroy (struct bstrList * sl) {
int i;
	if (sl == NULL) return BSTR_ERR;
	for (i=0; i < sl->qty; i++) {
		if (sl->entry[i]) {
			bdestroy (sl->entry[i]);
			sl->entry[i] = NULL;
		}
	}
	sl->qty = -1;
	free (sl);
	return BSTR_OK;
}

#include <stdarg.h>
#include <stdio.h>

/* Give WATCOM C/C++, MSVC some latitude for their non-support of vsnprintf */
#if defined(__WATCOMC__) || defined(_MSC_VER)
#define exvsnprintf(b,n,f,a) (_vsnprintf (b,n,f,a))
#else
#ifdef NOVSNP
/* This is just a hack.  If you are using a system without a vsnprintf, it is 
   not recommended that bformat be used at all. */
#define exvsnprintf(b,n,f,a) (vsprintf  (b,f,a))
#define START_VSNBUFF (256)
#else
#define exvsnprintf(b,n,f,a) (vsnprintf  (b,n,f,a))
#endif
#endif

#ifndef START_VSNBUFF
#define START_VSNBUFF (16)
#endif

/*  bstring bformat (const char * fmt, ...)
 *
 *  Takes the same parameters as printf (), but rather than outputting results
 *  to stdio, it forms a bstring which contains what would have been output.
 *  Note that if there is an early generation of a '\0' character, the 
 *  bstring will be truncated to this end point.
 */
bstring bformat (const char * fmt, ...) {
va_list arglist;
bstring buff;
int n;

	if (fmt == NULL) return NULL;

	/* Since the length is not determinable beforehand, a search is
	   performed using the truncating "vsnprintf" call (to avoid buffer
	   overflows) on increasing potential sizes for the output result. */

	buff = cstr2bstr ("");
	for (n = START_VSNBUFF; ; n += n) {
		if (BSTR_OK != balloc (buff, n + 2)) return NULL;

		va_start (arglist, fmt);
		exvsnprintf ((char *) buff->data, n + 1, fmt, arglist);
		va_end (arglist);

		buff->data[n] = '\0';
		buff->slen = (strlen) ((char *) buff->data);

		if (buff->slen < n) break;
	}

	return buff;
}

/*  bstring bvformat (const char * fmt, va_list arglist)
 *
 *  Takes the same parameters as vprintf (), but rather than outputting results
 *  to stdio, it forms a bstring which contains what would have been output.
 *  Note that if there is an early generation of a '\0' character, the 
 *  bstring will be truncated to this end point.
 */
bstring bvformat (const char * fmt, va_list arglist) {
bstring buff;
int n;

	if (fmt == NULL) return NULL;

	/* Since the length is not determinable beforehand, a search is
	   performed using the truncating "vsnprintf" call (to avoid buffer
	   overflows) on increasing potential sizes for the output result. */

	buff = cstr2bstr ("");
	for (n=16; ; n += n) {
		if (BSTR_OK != balloc (buff, n + 2)) return NULL;

		exvsnprintf ((char *) buff->data, n + 1, fmt, arglist);

		buff->data[n] = '\0';
		buff->slen = (strlen) ((char *) buff->data);

		if (buff->slen < n) break;
	}

	return buff;
}
