/*
 * This source file is part of the bstring string library.  This code was
 * written by Paul Hsieh in 2002-2003, and is covered by the BSD open source 
 * license. Refer to the accompanying documentation for details on usage and 
 * license.
 */

/*
 * bstrwrap.h
 *
 * This file is the C++ wrapper for the bstring functions.
 */

#include <malloc.h>
#include "bstrlib.h"

#ifdef __cplusplus

/////////////////// Configuration defines //////////////////////////////

// MSVC appears to just include STL automatically.  Other compilers,
// such as WATCOM C/C++ does not.
#if (defined(_MSC_VER) && !defined(BSTRLIB_CAN_USE_STL)) && !defined (BSTRLIB_CANNOT_USE_STL)
#define BSTRLIB_CAN_USE_STL
#endif

// MSVC and WATCOM C/C++ have options that make exception handling an
// optional feature.
#if !defined (BSTRLIB_THROWS_EXCEPTIONS) && !defined (BSTRLIB_DOESNT_THROW_EXCEPTIONS)
#define BSTRLIB_THROWS_EXCEPTIONS
#endif

////////////////////////////////////////////////////////////////////////

#ifdef BSTRLIB_THROWS_EXCEPTIONS
#define bstringThrow(er) throw ("CBString::" er)
#else
#define bstringThrow(er)
#endif

#if defined(BSTRLIB_CAN_USE_STL)
#include <vector>
#endif

struct CBString : public tagbstring {

	// Constructors
	CBString (char c) { *this = c; }
	CBString (unsigned char c) { *this = c; }
	CBString (int c) { *this = c; }
	CBString (float c) { *this = c; }
	CBString (double c) { *this = c; }
	CBString (const char *s = "") { *this = s; }
	CBString (const CBString& b) { *this = b; }
	CBString (const tagbstring& x) { *this = x; }
	CBString (char c, int len);
	CBString (const void * blk, int len);

	// Destructor
	~CBString ();

	// = operator
	const CBString& operator = (char c);
	const CBString& operator = (unsigned char c);
	const CBString& operator = (int c);
	const CBString& operator = (float c);
	const CBString& operator = (double c);
	const CBString& operator = (const char *s);
	const CBString& operator = (const CBString& b);
	const CBString& operator = (const tagbstring& x);

	// += operator
	const CBString& operator += (char c);
	const CBString& operator += (unsigned char c);
	const CBString& operator += (int c);
	const CBString& operator += (float c);
	const CBString& operator += (double c);
	const CBString& operator += (const char *s);
	const CBString& operator += (const CBString& b);

	// + operator
	const CBString operator + (char c);
	const CBString operator + (const CBString& b) const;
	const CBString operator + (const char *s) const;

	// Comparison operators
	bool operator == (const CBString& b) const;
	bool operator != (const CBString& b) const;
	bool operator <  (const CBString& b) const;
	bool operator <= (const CBString& b) const;	
	bool operator >  (const CBString& b) const;
	bool operator >= (const CBString& b) const;

	// Casts
	inline operator const char* () { return (const char *)data; }
	inline operator const unsigned char* () { return (const unsigned char *)data; }

	// Accessors
	inline int length () const {return slen;}
	inline unsigned char CBString::character (int i) const {
#ifdef BSTRLIB_THROWS_EXCEPTIONS
		if (((unsigned)i) >= (unsigned)slen) {
			bstringThrow ("character idx out of bounds");
		}
#endif
		return data[i];
	}
	inline unsigned char operator [] (int i) const { return character(i); }

	// Some useful string functions.
	void format (const char * fmt, ...);
	void vformat (const char * fmt, va_list arglist);
	int find (const CBString& b, int pos = 0) const;
	int reversefind (const CBString& b, int pos) const;
	int findchr (const CBString& b, int pos = 0) const;
	int reversefindchr (const CBString& b, int pos) const;
	const CBString midstr (int left, int len) const;
	void fill (int length, unsigned char fill = ' ');
	void alloc (int length);
	void setsubstr (int pos, const CBString& b, unsigned char fill = ' ');
	void insert (int pos, const CBString& b, unsigned char fill = ' ');
	void insertchrs (int pos, int len, unsigned char fill = ' ');
	void remove (int pos, int len);
	void trunc (int len);
	void repeat (int count);

	// Stream functions
	int gets (bNgetc getcPtr, void * parm, char terminator);
	int read (bNread readPtr, void * parm);

	// bStream functions
	int bStreamgets (struct bStream * s, char terminator);
	int bStreamread (struct bStream * s, int n);
	int bStreamunread (struct bStream * s);
	int bStreampeek (struct bStream * s);
};

#if defined(BSTRLIB_CAN_USE_STL)
struct CBStringList : public std::vector<CBString> {

	// split a string into a vector of strings.
	void split (const CBString& b, char splitChar);
	void split (const CBString& b, const CBString& s);
};
#endif
#endif
