#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "bstrlib.h"
#include "fctxt.h"
#include "glob2.h"

/* Command line switches */

static int optA = 0;
static int optB = 0;
static int optC = 0;
static int optN = 0;
static int optT = 0;
static int optW = 0;
static int optL = 0;

/*
 * ASCII comparison of a pair of lines, taking whitespace and case switches
 * into account.
 *
 * Returns 0 if and only if the two lines are the same.
 */
static int compareLine (const bstring b0, const bstring b1) {
int i, j, o0, o1;

    if (b0 == NULL || b1 == NULL) return -__LINE__;

    o0 = o1 = 0;
    for (i=j=0;;) {
        if (i >= b0->slen && j >= b1->slen) return 0;

        /* /W option -> virtually compresses all whitespaces to a single 
         * white space. */

        if (optW) {
            if (b0->data[i] == ' ' || b0->data[i] == '\t') {
                if (b1->data[j] != ' ' && b1->data[j] != '\t') return __LINE__;
                i++; j++;
                while (b0->data[i] == ' ' || b0->data[i] == '\t') i++;
                while (b1->data[j] == ' ' || b1->data[j] == '\t') j++;
            } else if (b0->data[i] == b1->data[j]) {
                i++; j++;
            } else {
                if (optC == 0) return __LINE__;
                if (toupper (b0->data[i]) != toupper (b1->data[j])) return -__LINE__;
                i++; j++;
            }

        /* /T option -> disable tab expansion */

        } else if (optT == 0) {
            if (b0->data[i] != '\t' && b1->data[j] != '\t') {
                if (b0->data[i] != b1->data[j]) {
                    if (optC == 0) return __LINE__;
                    if (toupper (b0->data[i]) != toupper (b1->data[j])) return -__LINE__;
                }
                i++; j++;
                o0++; o1++;
            } else {
                if (b0->data[i] == b1->data[j]) {
                    o0 = o1 = 0;
                }
                if (b0->data[i] != '\t') {
                    o0 = (o0 + 7) & ~7;
                    i++;
                    while (b1->data[j] == ' ') { 
                        j++; o1++; 
                        if (o0 == o1) goto nextChar;
                    }
                    if (b1->data[j] == '\t') { j++; o1 = (o1 + 7) & ~7; }
                    if (o0 != o1) return -__LINE__;
                }
                o1 = (o1 + 7) & ~7;
                j++;
                while (b0->data[i] == ' ') { 
                    i++; o0++; 
                    if (o0 == o1) goto nextChar;
                }
                if (b0->data[i] == '\t') { i++; o0 = (o0 + 7) & ~7; }
                if (o0 != o1) return -__LINE__;
            }
        } else {
            if (b0->data[i] != b1->data[j]) {
                if (optC == 0) return __LINE__;
                if (toupper (b0->data[i]) != toupper (b1->data[j])) return -__LINE__;
            }
            i++; j++;
        }

	nextChar:;
    }

    /* return -__LINE__; */
}

#define DEFAULT_MAX_MISMATCH (99)
#define DEFAULT_MIN_REMATCH (2)

static int maxMismatch = DEFAULT_MAX_MISMATCH;
static int minRematch = DEFAULT_MIN_REMATCH;

/*
 * Line buffer management functions.  These functions provide an API to the
 * line buffer window.
 */

struct lineBuff {
    int b, h, p, t;
    bstring * buff;
};

/* Constructor */
int lbInit (struct lineBuff * lb) {
int i;
    if (lb == NULL) return -__LINE__;
    lb->buff = (bstring *) malloc ((maxMismatch + minRematch) * sizeof (bstring));
    if (NULL == lb->buff) {
        printf ("FC: Out of memory\n");
        return -__LINE__;
    }
    for (i=0; i < maxMismatch + minRematch; i++) {
        lb->buff[i] = cstr2bstr ("");
    }
    lb->b = 0;
    lb->h = 0;
    lb->p = 0;
    lb->t = -1;
    return 0;
}

/* Destructor */
int lbDestroy (struct lineBuff * lb) {
int i;
    if (lb == NULL) return -__LINE__;
    for (i=0; i < maxMismatch + minRematch; i++) {
        bdestroy (lb->buff[i]);
        lb->buff[i] = NULL;
    }
    free (lb->buff);
    lb->buff = NULL;
    return 0;
}

/* How many entries are in the line buffer? */
int lbLength (struct lineBuff * lb) {
    if (lb == NULL) return -__LINE__;
    return (lb->h - lb->b + (maxMismatch + minRematch)) % (maxMismatch + minRematch);
}

/* Fetch the line indexed by the line number p.  If the window doesn't hold
 * that line, NULL is returned. */
bstring lbGet (struct lineBuff * lb, int p) {
int d;
    if (lb == NULL || p < 0) return NULL;
    p -= lb->p;
    if (p < -1) return NULL;
    d = lbLength (lb);
    if (p >= d) return NULL;
    p = (p + lb->b + (maxMismatch + minRematch)) % (maxMismatch + minRematch);
    return lb->buff[p];
}

/* Read in a line into line buffer window at the window tail. */
int lbRead (struct lineBuff * lb, FILE * fp) {
bstring b;
int ret;
    if (lb == NULL || fp == NULL || lb->t >= 0) return -__LINE__;
    if ((lb->h + 1) % (maxMismatch + minRematch) == lb->b) return -__LINE__;
    b = bgets ((bNgetc) fgetc, fp, '\n');
    if (b == NULL) {
        lb->t = lbLength (lb) + lb->p;
        return -__LINE__;
    }
    lb->buff[lb->h]->slen = 0;
    ret = bsetstr (lb->buff[lb->h], 0, b, '?');
    bdestroy (b);
    if (ret < 0) return -__LINE__;
    lb->h = (lb->h + 1) % (maxMismatch + minRematch);
    return 0;
}

/* Advance the window head d places. */
int lbAdvance (struct lineBuff * lb, int d) {
    if (d == 0) return 0;
    if (d < 0) return -__LINE__;
    if (d > lbLength (lb)) return -__LINE__;
    lb->b = (lb->b + d) % (maxMismatch + minRematch);
    lb->p += d;
    return 0;
}

/* Is the line essentially a blank line that the switches indicate that we
 * can ignore? */
int isIgnorableBlankLine (bstring b) {
int i;
    if (b == NULL || optW == 0) return 0;
    for (i=0; i < b->slen; i++) {
        if (b->data[i] == '\n') return 1;
        if (b->data[i] != '\r'
         && b->data[i] != ' '
         && b->data[i] != '\t') return 0;
    }
    return 1;
}

/* Display the lines from the buffer.  md is initially set to the last line
 * which has already been written, and is updated to the last line that is
 * written.  The /A and /N switches are dealt with here. */
int displayLines (int * md, struct lineBuff * lb, int p0, int p1) {
int k, s, ss, lOptA;
bstring b, t;

    if (lb == NULL || md == NULL) return -__LINE__;

    t = NULL;
    s = 0;
    if (p0 <= *md) p0 = *md + 1;

    lOptA = optA;
    if (p1-p0 < 3) lOptA = 0;

    for (k=p0; k <= p1; k++) {
        b = lbGet (lb, k);
        if (b == NULL) continue;
        if (isIgnorableBlankLine (b)) continue;
        *md = k;
        t = b;
        ss = s;
        s = k+1;
        if (ss != 0 && lOptA) continue;
        if (optN) printf ("%6d:  ", k+1);
        printf ("%s", b->data);
    }

    if (lOptA && s > 0) {
        printf ("...\n");
        if (optN) printf ("%6d:  ", s);
        printf ("%s", t->data);
    }

    return 0;
}

static int loptB;
static int loptL;

FILE * genFopen (const char * filename, const char * mode) {
char cName[272];
int r;

    if (filename == NULL || mode == NULL) return NULL;
    r = getShortFileNamePath (cName, filename);
    if (r == 0) {
        return fopen (cName, mode);
    }
    return fopen (filename, mode);
}

/* Determine whether or not the file is binary or ASCII text.  If no switch
 * has been provided at the command line, then up to the first 256 bytes of 
 * the file are examined to see if any non-ASCII characters are present.  If
 * the file is detected to be non-ASCII, then the global binary option flag
 * is set as if a /B switch were selected. */
static char * fileMode (const char * fileName) {
unsigned char buff[256];
FILE * fp;
int i, n;

    if (loptB) return "rb";
    if (loptL) return "r";
    fp = genFopen (fileName, "rb");
    if (fp == NULL) return "r";
    n = fread (buff, 1, 256, fp);
    fclose (fp);

    for (i=0; i < n; i++) {
        if ((buff[i] < 32 && buff[i] != 13 && buff[i] != 10 && buff[i] != 9 && buff[i] != 27) || (buff[i] > 127)) {
            loptB = 1;
            return "rb";
        }
    }
    return "r";
}

static char * fileName[2];
static int diffs;

/* Diff inner loop.  Basically when a mismatch is encountered, a pair of 
 * sub-windows are found such that the larger window is minimized, and the
 * two files match against right after these sub-windows.  This algorithm
 * yeilds results very similar to (though not always the same as) the 
 * Micrsoft FC implementation, allowing the two primary tweakable options 
 * (min rematch, and maximum mismatch) to behave similarly. Also handles
 * the /B switch here. */
int fcdiff (const char * file0, const char * file1) {
FILE * fp0, * fp1;
int l0, l1;

    loptB = optB;
    loptL = optL;

    fileMode (file0); fileMode (file1);

    fp0 = genFopen (file0, "rb");
    if (fp0 == NULL) {
	printf (fc_FILE_NOT_FOUND, file0);
	return -__LINE__;
    }

    fp1 = genFopen (file1, "rb");
    if (fp1 == NULL) {
	fclose (fp0);
	printf (fc_FILE_NOT_FOUND, file1);
	return -__LINE__;
    }

    printf (fc_COMPARING_FILES, file0, file1);

    diffs = 0;

    if (loptB) {
        unsigned char d0, d1;
	int p = 0;

        while (1) {
            l0 = fread (&d0, 1, 1, fp0);
            l1 = fread (&d1, 1, 1, fp1);

            if (l0 == 0 && l1 == 0) break;
            if (l0 == 0) {
                printf (fc_LENGTH_MISMATCH, file1, file0);
                break;
            }
            if (l1 == 0) {
                printf (fc_LENGTH_MISMATCH, file0, file1);
                break;
            }

	    if (d0 != d1) {
                printf ("%08X: %02X %02X\n", p, d0, d1);
            }
            p++;
	}
    } else {
	int f, p0, p1;
	int md0, md1;
        struct lineBuff lb0, lb1;

	if (lbInit (&lb0) || lbInit (&lb1)) {
            printf ("%s\n", fc_OUT_OF_MEMORY);
            return -__LINE__;
	}

	f = 0;
	p0 = p1 = 0;
        md0 = md1 = -1;

	while (f == 0) {
            int f0, f1, x0, x1;
            bstring b0, b1;

            f0 = lbRead (&lb0, fp0);
            f1 = lbRead (&lb1, fp1);

            if (p0 != lb0.p) {
                printf ("[%d] p0 = %d, lb0->p = %d\n", __LINE__, p0, lb0.p);
            }

            if (p1 != lb1.p) {
                printf ("[%d] p1 = %d, lb1->p = %d\n", __LINE__, p1, lb1.p);
            }

            b0 = lbGet (&lb0, p0);
            while (isIgnorableBlankLine (b0)) {
                lbAdvance (&lb0, 1);
                f0 = lbRead (&lb0, fp0);
                p0++;
                b0 = lbGet (&lb0, p0);
            }

            b1 = lbGet (&lb1, p1);
            while (isIgnorableBlankLine (b1)) {
                lbAdvance (&lb1, 1);
                f1 = lbRead (&lb1, fp1);
                p1++;
                b1 = lbGet (&lb1, p1);
            }

            x0 = lbLength (&lb0);
            x1 = lbLength (&lb1);

            if (x0 > 0 && x1 > 0) {
                if (0 == compareLine (b0, b1)) {
                    lbAdvance (&lb0, 1);
                    lbAdvance (&lb1, 1);
                    p0++;
                    p1++;
                } else {
                    int i, j, k;

                    diffs++;

                    /* Completely fill in the line buffer windows. */
                    while (lbLength (&lb0) <= (maxMismatch + minRematch) && 
                           lbLength (&lb1) <= (maxMismatch + minRematch)) {
                        f0 = lbRead (&lb0, fp0);
                        f1 = lbRead (&lb1, fp1);
                        if (f0 != 0 && f1 != 0) break;
                    }

                    /* Find (i,j) or (j,i) such that the windows rematch at
                     * their tails. */
                    for (i=1; i <= maxMismatch; i++) {
                        for (j=0; j <= i; j++) {
                            b0 = lbGet (&lb0, p0+i);
                            b1 = lbGet (&lb1, p1+j);
                            if ((md0+1 < p0 + i || md1+1 < p1 + j) && 0 == compareLine (b0, b1)) {
                                for (k = 1; k < minRematch; k++) {
                                    if (compareLine (lbGet (&lb0, p0+i+k), lbGet (&lb1, p1+j+k))) break;
                                }
                                if (k >= minRematch) {
                                    printf ("****** %s\n", file0);
                                    displayLines (&md0, &lb0, p0 - 1, p0 + i);
                                    printf ("****** %s\n", file1);
                                    displayLines (&md1, &lb1, p1 - 1, p1 + j);
                                    printf ("******\n\n");
                                    lbAdvance (&lb0, i);
                                    lbAdvance (&lb1, j);
                                    p0 += i;
                                    p1 += j;
                                    goto lineCont;
                                }
                            }

                            /* Don't need to check (j,i) if i == j */

                            if (j == i) continue;

                            b0 = lbGet (&lb0, p0+j);
                            b1 = lbGet (&lb1, p1+i);
                            if ((md0+1 < p0 + j || md1+1 < p1 + i) && 0 == compareLine (b0, b1)) {
                                for (k = 1; k < minRematch; k++) {
                                    if (compareLine (lbGet (&lb0, p0+j+k), lbGet (&lb1, p1+i+k))) break;
                                }
                                if (k >= minRematch) {
                                    printf ("****** %s\n", file0);
                                    displayLines (&md0, &lb0, p0 - 1, p0 + j);
                                    printf ("****** %s\n", file1);
                                    displayLines (&md1, &lb1, p1 - 1, p1 + i);
                                    printf ("******\n\n");
                                    lbAdvance (&lb0, j);
                                    lbAdvance (&lb1, i);
                                    p0 += j;
                                    p1 += i;
                                    goto lineCont;
                                }
                            }
                        }
                    }

                    x0 = lbLength (&lb0);
                    x1 = lbLength (&lb1);

                    /* A non-recoverable difference? */
                    if (x0 >= maxMismatch-1 || x1 >= maxMismatch-1)
                        printf ("%s\n", fc_RESYNCH_FAILED);

                    /* Either a non-recoverable difference, or the end of
                       both files. */
                    if (x0 > 0 || x1 > 0) {
                        printf ("****** %s\n", file0);
                        displayLines (&md0, &lb0, p0 - 1, p0 + x0);
                        printf ("****** %s\n", file1);
                        displayLines (&md1, &lb1, p1 - 1, p1 + x1);
                        printf ("******\n\n");
                    }

                    goto fcAbort;
                }
            } else {
                /* Has the EOF been reached in both files? */
                if (f0 != 0 && f1 != 0) {
                    if (x0 > 0 || x1 > 0) {
                        diffs++;
                        printf ("****** %s\n", file0);
                        displayLines (&md0, &lb0, p0 - 1, p0 + x0);
                        printf ("****** %s\n", file1);
                        displayLines (&md1, &lb1, p1 - 1, p1 + x1);
                        printf ("******\n\n");
                    }
                    break;
                }
            }

            /* ... */
            lineCont:;
	}

        fcAbort:;

        lbDestroy (&lb0);
        lbDestroy (&lb1);
    }

    fclose (fp0);
    fclose (fp1);

    if (diffs == 0) {
        printf ("%s", fc_NO_DIFFERENCES);
    }

    return 0;
}

/* Determine if the parameter has any wild cards in it. */

static int chkForWildCards (const char * pathName) {
int i;
    for (i=0; pathName[i]; i++) 
        if (pathName[i] == '*' || pathName[i] == '?') return 1;
    return 0;
}

/* Wrappers to the fcdiff function that can be plugged into the filename
 * walkers in the glob2 module. */

static int wrap01 (void * parm, const char * f) {
    fcdiff ((char *)parm, f);
    return 0;
}

static int wrap10 (void * parm, const char * f) {
    fcdiff (f, (char *)parm);
    return 0;
}

static int wrap11 (void * parm, const char * f0, const char * f1) {
    parm = parm;
    fcdiff (f0, f1);
    return 0;
}

/* Determine whether or not long filenames can be used, and whether or not
   wild card walking occurs */
int fcexec (char * file0, char * file1) {
int m0, m1, c;

    useLFN = (GetMSWin9XVer () >= 0x0400);

    m0 = chkForWildCards (file0);
    m1 = chkForWildCards (file1);
    if (m0 == 0) {
        if (m1 == 0) {
            return fcdiff (file0, file1);
        } else {
            c = onePathMasksWalkExpand (file1, wrap01, (void *) file0);
            if (c == 0) {
                printf (fc_FILE_NOT_FOUND, file1);
            }
        }
    } else {
        if (m1 == 0) {
            c = onePathMasksWalkExpand (file0, wrap10, (void *) file1);
            if (c == 0) {
                printf (fc_FILE_NOT_FOUND, file0);
            }
        } else {
            c = twoPathMasksWalkExpand (file0, file1, wrap11, NULL);
            if (c == 0) {
                printf (fc_FILE_NOT_FOUND, file1);
            }
        }
    }

    return (-(c <= 0)) & (-__LINE__);
}

/* Process command line switches, then call fcdiff */
int main (int argc, char * argv[]) {
int fnCount = 0;
int i, ret;

    for (i=1; i < argc; i++) {
	if (argv[i][0] == '/') {
	    switch (argv[i][1]) {
		case '?': {
	            printf ("%s\n", fc_TEXT_CMDHELP_DIR);
	            exit (0);
	        }
                case 'A': case 'a': optA = 1; break;
                case 'B': case 'b': optB = 1; break;
                case 'C': case 'c': optC = 1; break;
                case 'N': case 'n': optN = 1; break;
                case 'T': case 't': optT = 1; break;
                case 'W': case 'w': optW = 1; break;
		case 'L': case 'l': {
                    if (argv[i][2] == '\0') {
			optL = 1;
		    } else if (argv[i][2] == 'B' || argv[i][2] == 'b') {
			sscanf (argv[i]+3, "%d", &maxMismatch);
		    } else {
                        printf ("%s\n", fc_INVALID_SWITCH);
                    }
		    break;
		}
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9': {
		    sscanf (argv[i]+1, "%d", &maxMismatch);
		    break;
		}
		default: {
                    printf ("%s\n", fc_INVALID_SWITCH);
                }
            }
        } else {
            if (fnCount >= 2) {
                printf ("%s\n", fc_TOO_MANY_FILENAMES);
                exit (1);
            }
            fileName[fnCount] = argv[i];
            fnCount++;
	}
    }

    if (fnCount != 2) {
        printf ("%s", fc_NOT_ENOUGH_FILE_SPECS);
        exit (1);
    }

    if (optB & (optN|optL)) {
        printf ("%s", fc_INCOMPATIBLE_SWITCHES);
        exit (1);
    }

    ret = fcexec (fileName[0], fileName[1]);
    return 0;
}
